from collections import namedtuple
from copy import deepcopy

from drf_spectacular.contrib.django_filters import DjangoFilterExtension
from drf_spectacular.extensions import OpenApiFilterExtension
from rest_framework.filters import SearchFilter


class AudomaDjangoFilterExtension(DjangoFilterExtension):

    priority = 5

    def _get_x_choices(self, choices: namedtuple):
        return {"choices": {key: value for key, value in choices}}

    def resolve_filter_field(
        self, auto_schema, model, filterset_class, field_name, filter_field
    ):
        from audoma.plumbing import create_choices_enum_description

        choices = filter_field.extra.get("choices", None) or getattr(
            filter_field, "choices", None
        )
        if choices:
            # set proper help-text
            filter_field.extra["help_text"] = filter_field.extra.get(
                "help_text", "{choices}"
            )
            filter_field.extra["help_text"] = filter_field.extra["help_text"].format(
                choices=create_choices_enum_description(choices, field_name)
            )
        schemas = super().resolve_filter_field(
            auto_schema, model, filterset_class, field_name, filter_field
        )
        parsed_schemas = []
        for schema in schemas:
            if choices and "x-choices" not in schema:
                schema["schema"]["x-choices"] = self._get_x_choices(choices)
                parsed_schemas.append(schema)

        return parsed_schemas


class SearchFilterExtension(OpenApiFilterExtension):
    target_class = SearchFilter

    SEARCH_PARAMS = {
        "^": "Starts-with search.",
        "=": "Exact matches.",
        "@": "Full-text search.",
        "$": "Regex search.",
    }

    def _create_search_fields_description(self, fields):
        transformed_fields = []
        for field in fields:
            try:
                partials = field.split("__")
            except IndexError:
                transformed_fields.append(field)
                continue
            if len(partials) < 2:
                continue
            out = ""
            for x, partial in enumerate(partials[1:]):
                out += "\t " * (x + 1) + f"* `{partial}` \n"
            transformed_fields.append(out)
        return transformed_fields

    def _preprocess_fields(self, fields):
        new_fields = deepcopy(fields)
        for x, field in enumerate(fields):
            for keyword, description in self.SEARCH_PARAMS.items():
                if keyword in field:
                    new_fields[x] = field.replace(keyword, "")
                    new_fields[x] += f"({description})"
        return new_fields

    def _get_processed_search_fields(self, fields):
        transformed_fields = {}
        fields = self._preprocess_fields(fields)

        while fields:
            start_phrase = fields[0].split("__")[0]
            processed_fields = [f for f in fields if f.startswith(start_phrase)]
            transformed_fields[start_phrase] = self._create_search_fields_description(
                processed_fields
            )
            fields = [f for f in fields if f not in processed_fields]

        return transformed_fields

    def _get_custom_serach_filter_description(self, view):
        description = "Search by: \n"
        processed_fields = self._get_processed_search_fields(view.search_fields)
        for field_name, field_description in processed_fields.items():
            description += f"* `{field_name}` \n" + "".join(field_description)
        return description

    def get_schema_operation_parameters(self, schema):
        view = schema.view
        result = self.target.get_schema_operation_parameters(view)[0]
        result["description"] = result.get("description", "{search_fields}")
        # overwrite default description
        if result["description"] == self.target_class.search_description:
            result["description"] = "{search_fields}"

        result["description"] = result["description"].format(
            search_fields=self._get_custom_serach_filter_description(view)
        )
        return [result]



from unittest import TestCase

from drf_spectacular.plumbing import ComponentRegistry
from rest_framework import fields
from rest_framework.permissions import (
    BasePermission,
    IsAuthenticated,
)
from rest_framework.serializers import Serializer
from rest_framework.test import APIRequestFactory

from audoma.drf import (
    fields as audoma_fields,
    serializers as audoma_serializers,
)
from audoma.drf.validators import ExclusiveFieldsValidator
from audoma.openapi import AudomaAutoSchema
from audoma.tests.testtools import (
    create_basic_view,
    create_serializer,
    create_serializer_class,
    create_view_with_custom_action,
    create_view_with_custom_audoma_action,
)


class AudomaAutoSchemaTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.factory = APIRequestFactory()

    def test_get_description_no_permissions_success(self):
        view = create_basic_view(view_properties={"permission_classes": []})
        view.schema = AudomaAutoSchema()
        view.schema.method = "get"
        desc = view.schema.get_description()
        self.assertEqual(desc, """Test View Doc""")

    def test_get_description_with_basic_permissions_success(self):
        view = create_basic_view()
        view.schema = AudomaAutoSchema()
        view.schema.method = "get"
        desc = view.schema.get_description()
        self.assertIn("**Permissions:**", desc)
        self.assertIn("Test View Doc", desc)
        self.assertIn("`AllowAny`", desc)

    def test_get_description_with_custom_permission_success(self):
        class ExampleCustomPermission(BasePermission):
            """
            Some Custom Permission
            """

            def has_permission(self, requst, view):
                ...

        permission_classes = [ExampleCustomPermission & IsAuthenticated]
        view = create_basic_view(
            view_properties={"permission_classes": permission_classes}
        )
        view.schema = AudomaAutoSchema()
        view.schema.method = "get"
        desc = view.schema.get_description()
        self.assertIn("`ExampleCustomPermission`  &  `IsAuthenticated`", desc)
        self.assertIn("Test View Doc", desc)

    def test_get_description_no_description_permission_success(self):
        class SomePermission(BasePermission):
            ...

        permission_classes = [SomePermission]
        view = create_basic_view(
            view_properties={"permission_classes": permission_classes}
        )
        view.schema = AudomaAutoSchema()
        view.schema.method = "get"
        desc = view.schema.get_description()
        self.assertIn("(No description for this permission)", desc)

    def test_get_serializer_default_serializer_usage(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }

        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        view = create_basic_view(view_properties={"serializer_class": serializer_class})
        view.schema = AudomaAutoSchema()
        view.request = self.factory.get("/example/")
        view.action = "retrieve"
        serializer = view.schema._get_serializer()
        self.assertEqual(serializer_class, type(serializer))

    def test_get_serializer_collect_serializer_class_success(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }

        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        view = create_basic_view(
            view_properties={"common_collect_serializer_class": serializer_class}
        )
        view.schema = AudomaAutoSchema()
        view.request = self.factory.get("/example/")
        view.action = "retrieve"
        serializer = view.schema._get_serializer()
        self.assertEqual(serializer_class, type(serializer))

    def test_get_serializer_result_serializer_class_success(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }

        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        view = create_basic_view(
            view_properties={"common_result_serializer_class": serializer_class}
        )
        view.schema = AudomaAutoSchema()
        view.request = self.factory.get("/example/")
        view.action = "retrieve"
        serializer = view.schema._get_serializer(serializer_type="result")
        self.assertEqual(serializer_class, type(serializer))

    def test_get_serializer_retrieve_serializer_class_success(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }

        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        view = create_basic_view(
            view_properties={"retrieve_serializer_class": serializer_class}
        )
        view.schema = AudomaAutoSchema()
        view.request = self.factory.get("/example/")
        view.action = "retrieve"
        serializer = view.schema._get_serializer(serializer_type="result")
        self.assertEqual(serializer_class, type(serializer))

    def test_get_serializer_get_retrieve_serializer_class_success(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }

        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        view = create_basic_view(
            view_properties={"get_retrieve_serializer_class": serializer_class}
        )
        view.schema = AudomaAutoSchema()
        view.request = self.factory.get("/example/")
        view.action = "retrieve"
        serializer = view.schema._get_serializer(serializer_type="result")
        self.assertEqual(serializer_class, type(serializer))

    def test_get_serializer_get_retrieve_result_serializer_class_success(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }

        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        view = create_basic_view(
            view_properties={"get_retrieve_result_serializer_class": serializer_class}
        )
        view.schema = AudomaAutoSchema()
        view.request = self.factory.get("/example/")
        view.action = "retrieve"
        serializer = view.schema._get_serializer(serializer_type="result")
        self.assertEqual(serializer_class, type(serializer))

    def test_get_serializer_no_serializer_class_failure(self):
        view = create_basic_view(view_properties={"serializer_class": None})
        view.schema = AudomaAutoSchema()
        view.request = self.factory.get("/example/")
        view.action = "retrieve"
        serializer = view.schema._get_serializer(serializer_type="result")
        self.assertIsNone(serializer)

    def test_get_serializer_class_action_success(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }

        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )

        view = create_view_with_custom_action(
            action_name="custom_action",
            action_serializer_class=serializer_class,
            action_kwargs={"methods": ["GET"], "detail": False},
        )
        view.schema = AudomaAutoSchema()
        view.request = self.factory.get("/example/")
        view.action = "custom_action"
        serializer = view.schema._get_serializer(serializer_type="result")

        self.assertEqual(type(serializer), serializer_class)

    def test_get_serializer_class_audoma_action_success(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }

        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )

        view = create_view_with_custom_audoma_action(
            action_name="custom_action",
            action_serializer_class=serializer_class,
            action_kwargs={"methods": ["GET"], "detail": False},
        )
        view.schema = AudomaAutoSchema()
        view.request = self.factory.get("/example/")
        view.action = "custom_action"
        serializer = view.schema._get_serializer(serializer_type="result")

        self.assertEqual(type(serializer), serializer_class)

    def test_get_serializer_class_audoma_action_custom_responses_and_collectors(self):
        # response serializer
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
            "is_admin": fields.BooleanField(),
        }

        result_serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        # collect serializer

        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
        }

        collect_serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )

        request = self.factory.post("/example/", data={})

        view = create_view_with_custom_audoma_action(
            action_name="custom_action",
            action_kwargs={
                "methods": ["POST"],
                "detail": False,
                "collectors": {"post": collect_serializer_class},
                "results": {"post": {"201": result_serializer_class}},
            },
        )
        # in our case we have to call action to set some attrs

        view.request = request
        view.action = "custom_action"
        view.schema = AudomaAutoSchema()
        serializer = view.schema._get_serializer(serializer_type="result")

        self.assertEqual(type(serializer), dict)
        self.assertEqual(type(serializer["201"]), result_serializer_class)

        serializer = view.schema._get_serializer(serializer_type="collect")
        self.assertEqual(serializer, collect_serializer_class)

    def test_get_serializer_class_audoma_action_string_response_success(self):
        # response serializer
        request = self.factory.post("/example/", data={})

        view = create_view_with_custom_audoma_action(
            action_name="custom_action",
            action_kwargs={
                "methods": ["POST"],
                "detail": False,
                "results": "Test message",
            },
        )
        # in our case we have to call action to set some attrs

        view.request = request
        view.action = "custom_action"
        view.schema = AudomaAutoSchema()
        serializer = view.schema._get_serializer(serializer_type="result")

        self.assertEqual(type(serializer), dict)
        self.assertEqual(serializer["default"].description, "Test message")

    def test_map_serializer_field_charfield_success(self):
        fields_config = {
            "name": fields.CharField(max_length=255),
            "age": fields.IntegerField(),
            "is_admin": fields.BooleanField(),
        }

        serializer = create_serializer(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        serializer_fields = serializer.fields
        request = self.factory.get("/example/")
        view = create_basic_view(view_properties={"serializer_class": type(serializer)})

        view.request = request
        view.action = "retrieve"
        view.schema = AudomaAutoSchema()
        mapped_field = view.schema._map_serializer_field(
            serializer_fields["name"], direction="response"
        )
        self.assertEqual(type(mapped_field), dict)
        self.assertEqual(mapped_field["type"], "string")

    def test_map_serializer_field_choicefield_xchoices_success(self):
        fields_config = {
            "company_name": fields.CharField(max_length=255),
            "company_size": fields.ChoiceField(
                choices=(("SMALL", "small"), ("MEDIUM", "medium"), ("LARGE", "large"))
            ),
        }
        serializer = create_serializer(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        serializer_fields = serializer.fields

        request = self.factory.get("/example/")
        view = create_basic_view(view_properties={"serializer_class": type(serializer)})
        view.request = request
        view.action = "retrieve"
        view.schema = AudomaAutoSchema()
        mapped_field = view.schema._map_serializer_field(
            serializer_fields["company_size"], direction="response"
        )
        self.assertEqual(type(mapped_field), dict)
        self.assertEqual(mapped_field["type"], "string")
        self.assertEqual(type(mapped_field["x-choices"]), dict)
        self.assertEqual(
            mapped_field["x-choices"],
            {"choices": {"SMALL": "small", "MEDIUM": "medium", "LARGE": "large"}},
        )

    def test_map_serializer_field_audoma_choicefield_xchoices_success(self):
        fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "company_size": audoma_fields.ChoiceField(
                choices=(("SMALL", "small"), ("MEDIUM", "medium"), ("LARGE", "large"))
            ),
        }
        serializer = create_serializer(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        serializer_fields = serializer.fields

        request = self.factory.get("/example/")
        view = create_basic_view(view_properties={"serializer_class": type(serializer)})
        view.request = request
        view.action = "retrieve"
        view.schema = AudomaAutoSchema()
        mapped_field = view.schema._map_serializer_field(
            serializer_fields["company_size"], direction="response"
        )
        self.assertEqual(type(mapped_field), dict)
        self.assertEqual(mapped_field["type"], "string")
        self.assertEqual(type(mapped_field["x-choices"]), dict)
        self.assertEqual(
            mapped_field["x-choices"],
            {"choices": {"SMALL": "small", "MEDIUM": "medium", "LARGE": "large"}},
        )

    def test_get_request_for_media_type_is_bulk_create(self):
        fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "company_size": audoma_fields.CharField(),
        }
        serializer_class = create_serializer_class(
            fields_config=fields_config,
            serializer_base_classes=[
                Serializer,
                audoma_serializers.BulkSerializerMixin,
            ],
        )
        serializer = serializer_class()
        request = self.factory.post("/example/")
        view = create_basic_view(view_properties={"serializer_class": serializer_class})
        view.request = request
        view.action = "create"
        view.schema = AudomaAutoSchema()
        view.schema.registry = ComponentRegistry()
        view.schema.method = "POST"
        view.schema.is_bulk = True
        schema, _ = view.schema._get_request_for_media_type(serializer)
        self.assertEqual(type(schema), dict)
        self.assertEqual(type(schema["oneOf"]), list)
        self.assertEqual(type(schema["oneOf"][0]), dict)
        self.assertEqual(schema["oneOf"][0]["type"], "array")

    def test_get_request_for_media_type_is_bulk_update(self):
        fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "company_size": audoma_fields.CharField(),
        }
        serializer_class = create_serializer_class(
            fields_config=fields_config,
            serializer_base_classes=[
                Serializer,
                audoma_serializers.BulkSerializerMixin,
            ],
        )
        serializer = serializer_class()
        request = self.factory.put("/example/")
        view = create_basic_view(view_properties={"serializer_class": serializer_class})
        view.request = request
        view.action = "update"
        view.schema = AudomaAutoSchema()
        view.schema.registry = ComponentRegistry()
        view.schema.method = "PUT"
        view.schema.is_bulk = True
        schema, _ = view.schema._get_request_for_media_type(serializer)
        self.assertEqual(type(schema), dict)
        self.assertEqual(schema["type"], "array")

    def test_get_request_for_media_type_is_not_bulk(self):
        fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "company_size": audoma_fields.CharField(),
        }
        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        serializer = serializer_class()
        request = self.factory.post("/example/")
        view = create_basic_view(view_properties={"serializer_class": serializer_class})
        view.request = request
        view.action = "create"
        view.schema = AudomaAutoSchema()
        view.schema.registry = ComponentRegistry()
        view.schema.method = "POST"
        view.schema.is_bulk = False
        schema, _ = view.schema._get_request_for_media_type(serializer)
        self.assertEqual(type(schema), dict)
        self.assertEqual(schema["$ref"], "#/components/schemas/ExampleRequest")

    def test_map_serializer_default_serializer(self):
        fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "company_size": audoma_fields.CharField(),
        }
        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )
        serializer = serializer_class()
        request = self.factory.post("/example/")
        view = create_basic_view(view_properties={"serializer_class": serializer_class})
        view.request = request
        view.action = "create"
        view.schema = AudomaAutoSchema()
        view.schema.registry = ComponentRegistry()
        view.schema.method = "POST"
        view.schema.is_bulk = False
        schema = view.schema._map_serializer(serializer, direction="request")
        self.assertEqual(type(schema), dict)
        self.assertIn("company_name", schema["properties"])
        self.assertIn("company_size", schema["properties"])

    def test_map_serializer_with_exclusive_fields(self):
        fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "customer_name": audoma_fields.CharField(max_length=255),
        }
        serializer_class = create_serializer_class(
            fields_config=fields_config,
            serializer_base_classes=[Serializer],
            validators=[
                ExclusiveFieldsValidator(fields=["company_name", "customer_name"])
            ],
        )
        serializer = serializer_class()
        request = self.factory.post("/example/")
        view = create_basic_view(view_properties={"serializer_class": serializer_class})
        view.request = request
        view.action = "create"
        view.schema = AudomaAutoSchema()
        view.schema.registry = ComponentRegistry()
        view.schema.method = "POST"
        view.schema.is_bulk = False
        schema = view.schema._map_serializer(serializer, direction="request")
        self.assertEqual(type(schema["oneOf"]), list)
        self.assertIn("customer_name", schema["oneOf"][0]["properties"])
        self.assertNotIn("company_name", schema["oneOf"][0]["properties"])
        self.assertNotIn("customer_name", schema["oneOf"][1]["properties"])
        self.assertIn("company_name", schema["oneOf"][1]["properties"])

    def test_get_operation_id_bulk_create(self):
        fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "customer_name": audoma_fields.CharField(max_length=255),
        }
        serializer_class = create_serializer_class(
            fields_config=fields_config,
            serializer_base_classes=[Serializer],
            validators=[
                ExclusiveFieldsValidator(fields=["company_name", "customer_name"])
            ],
        )
        request = self.factory.post("/example/")
        view = create_basic_view(view_properties={"serializer_class": serializer_class})
        view.request = request
        view.action = "create"
        view.schema = AudomaAutoSchema()
        view.schema.registry = ComponentRegistry()
        view.schema.method = "POST"
        view.schema.is_bulk = True
        view.schema.path = "/example/"
        view.schema.path_prefix = ""
        view.schema.path_regex = r"\/\w\/"
        operation_id = view.schema.get_operation_id()
        self.assertIn("bulk", operation_id)

    def test_get_operation_id_bulk_upadte(self):
        fields_config = {
            "company_name": audoma_fields.CharField(max_length=255),
            "customer_name": audoma_fields.CharField(max_length=255),
        }
        serializer_class = create_serializer_class(
            fields_config=fields_config,
            serializer_base_classes=[Serializer],
            validators=[
                ExclusiveFieldsValidator(fields=["company_name", "customer_name"])
            ],
        )
        request = self.factory.post("/example/")
        view = create_basic_view(view_properties={"serializer_class": serializer_class})
        view.request = request
        view.action = "create"
        view.schema = AudomaAutoSchema()
        view.schema.registry = ComponentRegistry()
        view.schema.method = "POST"
        view.schema.is_bulk = False
        view.schema.path = "/example/"
        view.schema.path_prefix = ""
        view.schema.path_regex = r"\/\w\/"
        operation_id = view.schema.get_operation_id()
        self.assertNotIn("bulk", operation_id)



from datetime import date

from audoma_api.models import (
    Car,
    CarTag,
    ExampleFileModel,
    ExampleModel,
    ExamplePerson,
    Manufacturer,
)

from audoma.choices import make_choices
from audoma.drf import serializers
from audoma.drf.decorators import document_and_format
from audoma.drf.serializers import (
    BulkListSerializer,
    BulkSerializerMixin,
)
from audoma.drf.validators import ExclusiveFieldsValidator


class NestedExampleSerializer(serializers.Serializer):
    email = serializers.EmailField()


class ExampleSerializer(serializers.Serializer):
    charfield_nolimits = serializers.CharField()
    charfield_min_max = serializers.CharField(min_length=10, max_length=20)
    phone_number = serializers.PhoneNumberField()
    phone_number_example = serializers.PhoneNumberField(example="+48 123 456 789")
    email = serializers.EmailField()
    url = serializers.URLField()
    boolean = serializers.BooleanField()
    nullboolean = serializers.NullBooleanField()
    mac_address = serializers.MACAddressField()
    regex_mac_address = serializers.RegexField(
        regex="^([0-9A-F]{2}:){5}([0-9A-F]{2})|([0-9A-F]{2}-){5}([0-9A-F]{2})$"
    )
    slug = serializers.SlugField()
    uuid = serializers.UUIDField(format="hex")
    # file_path = serializers.FilePathField()
    ip_address = serializers.IPAddressField()
    integer = serializers.IntegerField(min_value=25, max_value=30)
    float = serializers.FloatField(min_value=1.0, max_value=4.0)
    decimal = serializers.DecimalField(max_digits=10, decimal_places=2)
    datetime = serializers.DateTimeField()
    date = serializers.DateField(example=str(date.today()))
    time = serializers.TimeField(example="12:34:56.000000")
    duration = serializers.DurationField()
    money = serializers.MoneyField(max_digits=10, decimal_places=2)
    choice = serializers.ChoiceField({1: "One", 2: "Two", 3: "Three"})
    # multiple_choice = serializers.MultipleChoiceField({1: "One", 2: "Two", 3: "Three"})
    list_of_emails = serializers.ListField(child=serializers.EmailField())
    dict_of_addressess = serializers.DictField(child=serializers.EmailField())
    hstore_of_emails = serializers.HStoreField(child=serializers.EmailField())
    json = serializers.JSONField()
    readonly = serializers.ReadOnlyField()
    nested = NestedExampleSerializer()

    class Meta:
        model = ExampleModel


class ExampleModelSerializer(serializers.ModelSerializer):
    phone_number = serializers.SerializerMethodField()

    class Meta:

        model = ExampleModel
        fields = "__all__"
        extra_kwargs = {"char_field": {"example": "lorem ipsum"}}

    @document_and_format(serializers.PhoneNumberField)
    def get_phone_number(self, obj):
        return obj.phone_number


class ExamplePersonModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExamplePerson
        fields = "__all__"


class ExampleFileModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExampleFileModel
        fields = "__all__"


class ManufacturerModelSerializer(BulkSerializerMixin, serializers.ModelSerializer):
    class Meta:
        model = Manufacturer
        list_serializer_class = BulkListSerializer
        fields = "__all__"


class CarModelSerializer(serializers.ModelSerializer):

    choices_options_links = {
        "manufacturer": {
            "viewname": "manufacturer_viewset-list",
            "value_field": "id",
            "display_field": "name",
        }
    }

    manufacturer = serializers.IntegerField()

    class Meta:
        model = Car
        fields = ["name", "body_type", "manufacturer", "engine_size", "engine_type"]


class CarTagSerializer(serializers.ModelSerializer):
    class Meta:
        model = CarTag
        fields = ["name"]


class CarDetailModelSerializer(serializers.ModelSerializer):
    manufacturer_name = serializers.CharField(
        read_only=True, source="manufacturer__name"
    )

    tags = CarTagSerializer(many=True, required=False)

    class Meta:
        model = Car
        fields = [
            "id",
            "name",
            "body_type",
            "manufacturer",
            "engine_size",
            "engine_type",
            "tags",
            "manufacturer_name",
        ]

    def _handle_tags(self, tags, instance):
        names = []
        for tag in tags:
            if instance.tags.filter(name=tag["name"]).exists():
                continue
            serializer = CarTagSerializer(data=tag)
            serializer.is_valid(raise_exception=True)
            serializer.save(car_id=instance.id)
            names.append(tag["name"])

        if tags is not None:
            CarTag.objects.exclude(name__in=names).delete()

    def update(self, instance, validated_data):
        tags = validated_data.pop("tags")
        self._handle_tags(tags, instance)
        return super().update(instance, validated_data)

    def create(self, validated_data):
        tags = validated_data.pop("tags")
        instance = super().create(validated_data)
        self._handle_tags(tags, instance)
        instance.refresh_from_db()
        return instance


class MutuallyExclusiveExampleSerializer(serializers.Serializer):
    class Meta:
        validators = [
            ExclusiveFieldsValidator(
                fields=[
                    "example_field",
                    "second_example_field",
                ]
            ),
            ExclusiveFieldsValidator(
                fields=[
                    "third_example_field",
                    "fourth_example_field",
                ]
            ),
        ]

    example_field = serializers.CharField(required=False)
    second_example_field = serializers.CharField(required=False)
    third_example_field = serializers.CharField(required=False)
    fourth_example_field = serializers.CharField(required=False)

    not_exclusive_field = serializers.CharField(required=False)
    second_not_exclusive_field = serializers.CharField(required=False)


class ExampleModelCreateSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExampleModel
        fields = "__all__"

    def update(self):
        for key, item in self.validated_data.items():
            setattr(self.instance, key, item)
        return self.instance

    def create(self):
        return ExampleModel(
            char_field=self.validated_data["char_field"],
            phone_number=self.validated_data["phone_number"],
            email=self.validated_data["email"],
            url=self.validated_data["url"],
            boolean=self.validated_data["boolean"],
            nullboolean=self.validated_data["nullboolean"],
            mac_adress=self.validated_data["mac_adress"],
            slug=self.validated_data["slug"],
            uuid=self.validated_data["uuid"],
            ip_address=self.validated_data["ip_address"],
            integer=self.validated_data["integer"],
            _float=self.validated_data["_float"],
            decimal=self.validated_data["decimal"],
            datetime=self.validated_data["datetime"],
            date=self.validated_data["date"],
            time=self.validated_data["time"],
            duration=self.validated_data["duration"],
            choices=self.validated_data["choices"],
            json=self.validated_data["json"],
        )

    def save(self, **kwargs):
        if self.instance:
            return self.update()
        return self.create()


class ExampleOneFieldSerializer(serializers.Serializer):

    RATES = make_choices("RATE", ((0, "LIKE", "Like"), (1, "DISLIKIE", "Dislike")))

    rate = serializers.ChoiceField(choices=RATES.get_choices())

    def save(self, **kwargs):
        return self.data



# Generated by Django 3.2.12 on 2022-06-20 10:14

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)

import audoma.django.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ("audoma_api", "0002_alter_examplemodel_json_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="Manufacturer",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", audoma.django.db.fields.CharField(max_length=255)),
                ("slug_name", audoma.django.db.fields.SlugField()),
            ],
        ),
        migrations.CreateModel(
            name="Car",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", audoma.django.db.fields.CharField(max_length=255)),
                (
                    "body_type",
                    audoma.django.db.fields.IntegerField(
                        choices=[
                            (1, "Sedan"),
                            (2, "Coupe"),
                            (3, "Hatchback"),
                            (4, "Pickup Truck"),
                        ]
                    ),
                ),
                ("engine_size", audoma.django.db.fields.FloatField()),
                (
                    "engine_type",
                    audoma.django.db.fields.IntegerField(
                        choices=[
                            (1, "Petrol"),
                            (2, "Diesel"),
                            (3, "Electric"),
                            (4, "Hybrid"),
                        ]
                    ),
                ),
                (
                    "manufacturer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="audoma_api.manufacturer",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.0.4 on 2022-05-10 14:09

from django.db import migrations

import audoma.django.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ("audoma_api", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="examplemodel",
            name="json",
            field=audoma.django.db.fields.JSONField(),
        ),
        migrations.AlterField(
            model_name="examplemodel",
            name="mac_adress",
            field=audoma.django.db.fields.MACAddressField(integer=True),
        ),
        migrations.AlterField(
            model_name="examplemodel",
            name="money_currency",
            field=audoma.django.db.fields.CurrencyField(
                choices=[
                    ("AUD", "Australian Dollar"),
                    ("GBP", "British Pound"),
                    ("CNY", "Chinese Yuan"),
                    ("EUR", "Euro"),
                    ("INR", "Indian Rupee"),
                    ("JPY", "Japanese Yen"),
                    ("NZD", "New Zealand Dollar"),
                    ("CHF", "Swiss Franc"),
                    ("USD", "US Dollar"),
                ],
                default="XYZ",
                max_length=3,
            ),
        ),
        migrations.AlterField(
            model_name="exampleperson",
            name="savings_currency",
            field=audoma.django.db.fields.CurrencyField(
                choices=[
                    ("AUD", "Australian Dollar"),
                    ("GBP", "British Pound"),
                    ("CNY", "Chinese Yuan"),
                    ("EUR", "Euro"),
                    ("INR", "Indian Rupee"),
                    ("JPY", "Japanese Yen"),
                    ("NZD", "New Zealand Dollar"),
                    ("CHF", "Swiss Franc"),
                    ("USD", "US Dollar"),
                ],
                default="PLN",
                max_length=3,
            ),
        ),
    ]



# Generated by Django 3.2.12 on 2022-04-21 12:41

from django.db import (
    migrations,
    models,
)

import audoma.django.db.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ExampleFileModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("file_field", models.FileField(upload_to="")),
                ("name", audoma.django.db.fields.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="ExampleModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("char_field", audoma.django.db.fields.CharField(max_length=255)),
                (
                    "phone_number",
                    audoma.django.db.fields.PhoneNumberField(
                        max_length=128, region=None
                    ),
                ),
                (
                    "phone_number_example",
                    audoma.django.db.fields.PhoneNumberField(
                        max_length=128, region=None
                    ),
                ),
                ("email", audoma.django.db.fields.EmailField(max_length=254)),
                ("text_field", audoma.django.db.fields.TextField()),
                ("url", audoma.django.db.fields.URLField()),
                ("boolean", audoma.django.db.fields.BooleanField()),
                ("nullboolean", audoma.django.db.fields.BooleanField(null=True)),
                ("mac_adress", audoma.django.db.fields.MACAddressField(integer=True)),
                ("slug", audoma.django.db.fields.SlugField()),
                ("uuid", audoma.django.db.fields.UUIDField()),
                ("ip_address", audoma.django.db.fields.GenericIPAddressField()),
                ("integer", audoma.django.db.fields.IntegerField()),
                ("_float", audoma.django.db.fields.FloatField()),
                (
                    "decimal",
                    audoma.django.db.fields.DecimalField(
                        decimal_places=2, max_digits=10
                    ),
                ),
                ("datetime", audoma.django.db.fields.DateTimeField()),
                ("date", audoma.django.db.fields.DateField()),
                ("time", audoma.django.db.fields.TimeField()),
                ("duration", audoma.django.db.fields.DurationField()),
                (
                    "choices",
                    audoma.django.db.fields.IntegerField(
                        choices=[(1, "example 1"), (2, "example 2"), (3, "example 3")]
                    ),
                ),
                ("json", audoma.django.db.fields.JSONField(default=dict)),
                (
                    "money_currency",
                    audoma.django.db.fields.CurrencyField(
                        choices=[("EUR", "Euro"), ("USD", "US Dollar")],
                        default="XYZ",
                        max_length=3,
                    ),
                ),
                (
                    "money",
                    audoma.django.db.fields.MoneyField(decimal_places=2, max_digits=10),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ExamplePerson",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "first_name",
                    audoma.django.db.fields.CharField(max_length=225, default="Adam"),
                ),
                ("last_name", audoma.django.db.fields.CharField(max_length=255)),
                ("age", audoma.django.db.fields.IntegerField()),
                ("email", audoma.django.db.fields.EmailField(max_length=254)),
                ("birth_country", audoma.django.db.fields.CharField(max_length=255)),
                ("residence_city", audoma.django.db.fields.CharField(max_length=255)),
                ("has_valid_account", audoma.django.db.fields.BooleanField()),
                ("ip_address", audoma.django.db.fields.GenericIPAddressField()),
                (
                    "savings_currency",
                    audoma.django.db.fields.CurrencyField(
                        choices=[("EUR", "Euro"), ("USD", "US Dollar")],
                        default="PLN",
                        max_length=3,
                    ),
                ),
                (
                    "savings",
                    audoma.django.db.fields.MoneyField(
                        decimal_places=2, default_currency="PLN", max_digits=14
                    ),
                ),
                (
                    "phone_number",
                    audoma.django.db.fields.PhoneNumberField(
                        max_length=128, region="IT"
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.1.2 on 2022-10-26 08:32

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)

import audoma.django.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ("audoma_api", "0005_auto_20220706_1013"),
    ]

    operations = [
        migrations.CreateModel(
            name="CarTag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", audoma.django.db.fields.CharField(max_length=255)),
                (
                    "car",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tags",
                        to="audoma_api.car",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.0.4 on 2022-05-26 14:39

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("audoma_api", "0002_alter_examplemodel_json_and_more"),
        ("audoma_api", "0002_examplesimplemodel_alter_examplemodel_json_and_more"),
    ]

    operations = []



from typing import NamedTuple

from django_filters import rest_framework as df_filters

from audoma.plumbing import create_choices_enum_description


class DocumentedTypedChoiceFilter(df_filters.TypedChoiceFilter):
    """Extended TypedChoiceFilter to generate documentation automatically"""

    def __init__(self, full_choices: NamedTuple, parameter_name: str, **kwargs) -> None:
        super().__init__(
            coerce=lambda value: full_choices.get_value_by_name(value),
            choices=full_choices.get_api_choices(),
            **kwargs,
        )
        self.full_choices = full_choices
        self.parameter_name = parameter_name
        self.extra["help_text"] = self.extra.get("help_text", "{choices}").format(
            choices=create_choices_enum_description(
                full_choices.get_api_choices(), self.field_name
            )
        )
        self.extra["choices"] = full_choices.get_api_choices()



from typing import (
    Callable,
    Type,
)

from rest_framework import generics
from rest_framework.settings import api_settings

from audoma.decorators import AudomaArgs
from audoma.drf.serializers import (
    BaseSerializer,
    DefaultMessageSerializer,
)
from audoma.operations import OperationExtractor


LOCATION_HEADER_STATUSES = [201] + list(range(300, 400))


class GenericAPIView(generics.GenericAPIView):

    """
    Extended GenericAPIView known from rest_framework.
    This class extends `get_serializer` and `get_serializer_class` methods.
    Also provides `get_result_serializer`, which is a shourtcut for `get_serializer` with proper param.
    """

    def get_serializer(self, *args, **kwargs) -> BaseSerializer:
        """
        Passes additional param to `get_serializer_class`.

        kwargs:
            serializer_type - defines if serializer is collect or result serializer.
                result serializer will be used to produce response, collect to process incoming data.
            serializer_class - it is possible to pass serializer_class to get_serializer, this will
                ends with returning passed serializer_class object.

        Returns:
            Object of obtained serializer class.
        """
        many = kwargs.get("many", False)

        serializer_type = kwargs.pop("serializer_type", "collect")
        status_code = kwargs.pop("status_code", None)
        serializer_class = kwargs.pop(
            "serializer_class",
            self.get_serializer_class(
                serializer_type=serializer_type, many=many, status_code=status_code
            ),
        )
        kwargs["context"] = self.get_serializer_context()

        if (
            kwargs.get("instance") is None
            and serializer_class is DefaultMessageSerializer
        ):
            audoma_args = self.get_audoma_action_config()
            message = self._extract_audoma_action_serializer(
                serializer_type, audoma_args, status_code
            )
            kwargs["instance"] = {"message": message}

        return serializer_class(*args, **kwargs)

    # needed by AudomaSwaggerAutoSchema
    def get_result_serializer(self, *args, **kwargs) -> BaseSerializer:
        """
        Shortuct for get_serializer.
        Simply has serializer_type set to `result`
        """
        return self.get_serializer(*args, serializer_type="result", **kwargs)

    def _check_action_function(self):
        func = getattr(self, self.action, None)
        if func and callable(func):
            return hasattr(func, "_audoma")
        return False

    def _extract_audoma_action_serializer(
        self, serializer_type: str, audoma_args: AudomaArgs, status_code: int = None
    ):

        extractor = OperationExtractor(
            audoma_args.collectors, audoma_args.results, audoma_args.errors
        )
        operation_category = "response" if serializer_type == "result" else "collect"

        # TODO - consider this idea
        if not status_code and operation_category == "response":
            return None

        return extractor.extract_operation(
            self.request, status_code, operation_category
        )

    def get_audoma_action_config(self):
        audoma_args = None
        if self._check_action_function():
            func = getattr(self, self.action)
            audoma_args = getattr(func, "_audoma", None)
        return audoma_args

    def get_audoma_action_serializer_class(
        self, serializer_type: str, status_code: int = None
    ):
        audoma_args = self.get_audoma_action_config()
        if audoma_args is None:
            return None

        serializer_class = self._extract_audoma_action_serializer(
            serializer_type, audoma_args, status_code
        )

        if isinstance(serializer_class, str):
            serializer_class = DefaultMessageSerializer

        return serializer_class

    def get_serializer_class(
        self,
        serializer_type: str = "collect",
        many: bool = False,
        status_code: int = None,
    ) -> Type[BaseSerializer]:
        """
        Extends defuault `get_serializer_class` method.
        This returns proper serializer_class for current request.

        Args:
            serializer_type - serializer_type of serializer to be returned, it may be collect or result serializer.

        Returns:
            This returns serializer_class

        """
        assert self.action not in [
            "post",
            "put",
            "delete",
            "patch",
            "options",
            "get",
            "head",
        ]
        serializer_class = self.get_audoma_action_serializer_class(
            serializer_type, status_code
        )
        if not serializer_class:
            method = self.request.method.lower()
            if self.action == "metadata":
                action = self.action_map.get("post", "list")
            else:
                action = self.action
            attr_names = [
                "%s_%s_%s_serializer_class" % (method, action, serializer_type),
                "%s_%s_serializer_class" % (action, serializer_type),
                "%s_%s_serializer_class" % (method, action),
                "%s_serializer_class" % (action),
                "common_%s_serializer_class" % (serializer_type),
                "serializer_class",
            ]
            for attr_name in attr_names:
                try:
                    serializer_class = getattr(self, attr_name)
                except AttributeError:
                    continue
                else:
                    break

            assert serializer_class is not None, (
                "'%s' should either include a `serializer_class`  attribute, "
                "or override the `get_serializer_class()` method."
                % self.__class__.__name__
            )

        if (
            serializer_type == "result"
            and self.action != "list"
            and hasattr(serializer_class, "get_result_serializer_class")
        ):
            assert callable(serializer_class.get_result_serializer_class)
            serializer_class = serializer_class.get_result_serializer_class()

        return serializer_class

    def _retrieve_response_headers(
        self,
        status_code: int,
        result_serializer: BaseSerializer,
    ):
        method = self.request.method.lower()

        method_names = [
            f"get_{self.action}_{self.request.method.lower()}_response_headers",
            f"get_{self.action}_response_headers",
            f"get_{self.request.method}_response_headers",
            "get_response_headers",
        ]
        headers = {}
        for name in method_names:
            method = getattr(self, name, None)
            if not method or not isinstance(method, Callable):
                continue
            headers = method(result_serializer)
            break

        # Append Location header for status proper status codes automatically.
        # For more info check:
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Location
        if status_code in LOCATION_HEADER_STATUSES:
            data = result_serializer.data
            try:
                if "Location" not in headers:
                    headers["Location"] = str(data[api_settings.URL_FIELD_NAME])
            except KeyError:
                pass
        return headers



from rest_framework import serializers
from rest_framework.test import APIRequestFactory

from django.test import TestCase

from audoma.tests import testtools


class GenericAPIViewTestCase(TestCase):
    def setUp(self):
        serializer_base_classes = (serializers.Serializer,)
        self.result_serializer_class = testtools.create_serializer_class(
            fields_config={"name": serializers.CharField(max_length=255)},
            serializer_base_classes=serializer_base_classes,
        )
        self.collect_serializer_class = testtools.create_serializer_class(
            fields_config={
                "name": serializers.CharField(max_length=255),
                "created_by": serializers.CharField(max_length=255),
            },
            serializer_base_classes=serializer_base_classes,
        )
        self.factory = APIRequestFactory()

    def test_get_serializer_result_success(self):
        viewset = testtools.create_basic_view(
            view_properties={
                "common_result_serializer_class": self.result_serializer_class
            },
        )
        viewset.action = "create"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_result_serializer()
        self.assertEqual(type(serializer), self.result_serializer_class)

    def test_get_serializer_collect_success(self):
        viewset = testtools.create_basic_view(
            view_properties={
                "common_collect_serializer_class": self.collect_serializer_class
            },
        )
        viewset.action = "create"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_serializer(serializer_type="collect")
        self.assertEqual(type(serializer), self.collect_serializer_class)

    def test_get_serializer_default_success(self):
        viewset = testtools.create_basic_view(
            view_properties={"serializer_class": self.result_serializer_class},
        )
        viewset.action = "create"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_serializer(serializer_type="result")
        self.assertEqual(type(serializer), self.result_serializer_class)

    def test_get_serializer_result_many_success(self):
        viewset = testtools.create_basic_view(
            view_properties={
                "common_result_serializer_class": self.result_serializer_class
            },
        )
        viewset.action = "create"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_result_serializer(many=True)
        self.assertEqual(type(serializer), serializers.ListSerializer)

    def test_get_serializer_collect_many_success(self):
        viewset = testtools.create_basic_view(
            view_properties={
                "common_collect_serializer_class": self.result_serializer_class
            },
        )
        viewset.action = "create"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_serializer(many=True)
        self.assertEqual(type(serializer), serializers.ListSerializer)

    def test_get_serializer_class_audoma_action_result_success(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            view_properties={"serializer_class": self.result_serializer_class},
            action_kwargs={
                "collectors": self.collect_serializer_class,
                "results": self.result_serializer_class,
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 201),
            },
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_result_serializer()
        self.assertEqual(type(serializer), self.result_serializer_class)

    def test_get_serializer_class_audoma_action_collect_success(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            view_properties={"serializer_class": self.result_serializer_class},
            action_kwargs={
                "collectors": self.collect_serializer_class,
                "results": self.result_serializer_class,
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 201),
            },
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_serializer()
        self.assertEqual(type(serializer), self.collect_serializer_class)

    def test_get_serializer_class_audoma_action_result_method_nested_success(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            view_properties={"serializer_class": self.result_serializer_class},
            action_kwargs={
                "collectors": self.collect_serializer_class,
                "results": {"post": self.result_serializer_class},
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 201),
            },
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_result_serializer()
        self.assertEqual(type(serializer), self.result_serializer_class)

    def test_get_serializer_class_audoma_action_collect_method_nested_success(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            view_properties={"serializer_class": self.result_serializer_class},
            action_kwargs={
                "collectors": {"post": self.collect_serializer_class},
                "results": self.result_serializer_class,
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 201),
            },
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_serializer()
        self.assertEqual(type(serializer), self.collect_serializer_class)

    def test_get_serializer_class_audoma_action_result_method_code_nested_success(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            view_properties={"serializer_class": self.result_serializer_class},
            action_kwargs={
                "collectors": self.collect_serializer_class,
                "results": {"post": {201: self.result_serializer_class}},
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 201),
            },
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        serializer = viewset.get_result_serializer()
        self.assertEqual(type(serializer), self.result_serializer_class)

    def test_get_serializer_class_audoma_action_result_method_code_nested_failure(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            action_kwargs={
                "collectors": self.collect_serializer_class,
                "results": {"post": {201: self.result_serializer_class}},
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 200),
            }
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        try:
            viewset.get_result_serializer()
        except Exception as e:
            self.assertEquals(type(e), AssertionError)
            self.assertEquals(
                str(e),
                "'ExampleView' should either include a `serializer_class` "
                + " attribute, or override the `get_serializer_class()` method.",
            )

    def test_retrieve_response_headers_for_action_and_method_success(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            action_kwargs={
                "collectors": self.collect_serializer_class,
                "results": {"post": {201: self.result_serializer_class}},
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 201),
            }
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        viewset.get_custom_action_post_response_headers = lambda x: {
            "CustomHeader": "X"
        }
        headers = viewset._retrieve_response_headers(
            201, self.result_serializer_class()
        )
        self.assertDictEqual(headers, {"CustomHeader": "X"})

    def test_retrieve_response_headers_for_action_success(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            action_kwargs={
                "collectors": self.collect_serializer_class,
                "results": {"post": {201: self.result_serializer_class}},
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 201),
            }
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        viewset.get_custom_action_response_headers = lambda x: {"CustomHeader": "X"}
        headers = viewset._retrieve_response_headers(
            201, self.result_serializer_class()
        )
        self.assertDictEqual(headers, {"CustomHeader": "X"})

    def test_retrieve_common_response_headers_success(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            action_kwargs={
                "collectors": self.collect_serializer_class,
                "results": {"post": {201: self.result_serializer_class}},
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 201),
            }
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        viewset.get_response_headers = lambda x: {"CustomHeader": "X"}
        headers = viewset._retrieve_response_headers(
            201, self.result_serializer_class()
        )
        self.assertDictEqual(headers, {"CustomHeader": "X"})

    def test_retrieve_response_headers_fallback_success(self):
        viewset = testtools.create_view_with_custom_audoma_action(
            action_kwargs={
                "collectors": self.collect_serializer_class,
                "results": {"post": {201: self.result_serializer_class}},
                "detail": False,
                "methods": ["post"],
                "returnables": ({}, 201),
            }
        )
        viewset.action = "custom_action"
        viewset.format_kwarg = "json"
        viewset.request = self.factory.post("/example/")
        headers = viewset._retrieve_response_headers(
            201, self.result_serializer_class()
        )
        self.assertDictEqual(headers, {})



"""Audoma Django model Fields
This module contains all the fields from Django models with additional functionality.
By inheriting from Audoma Mixins, an example is generated for each field (i.e. FloatField will have
example generated based on field's min and max values).
We can define custom example by simply passing `example` as an argument to the field.
"""

import random
import sys

import phonenumbers
from djmoney.models import fields as djmoney_fields
from djmoney.utils import get_currency_field_name
from macaddress.fields import MACAddressField
from phonenumber_field.modelfields import PhoneNumberField
from phonenumber_field.phonenumber import to_python

from django.db import models
from django.db.models.fields import (  # noqa: F401
    BLANK_CHOICE_DASH,
    NOT_PROVIDED,
    __all__,
)

from audoma.django import forms
from audoma.example_generators import generate_lorem_ipsum
from audoma.mixins import ModelExampleMixin


try:
    from django.db.models import JSONField
except ImportError:
    try:
        from jsonfield import JSONField
    except ImportError as err:
        raise ImportError(
            "You are using old version of Django that doesn't support JSONField. Please install django-jsonfield"
        ) from err


this = sys.modules[__name__]


for field_name in __all__:
    if field_name in ["BLANK_CHOICE_DASH", "NOT_PROVIDED"]:
        continue

    setattr(
        this,
        field_name,
        type(field_name, (ModelExampleMixin, getattr(models, field_name)), {}),
    )


__all__.extend(["MoneyField", "CurrencyField", "PhoneNumberField", "MACAddressField"])


class CurrencyField(ModelExampleMixin, djmoney_fields.CurrencyField):
    def __init__(self, *args, **kwargs) -> None:
        default = kwargs.get("default", None)
        if default and str(default) != "XYZ":
            self.example = default
        elif kwargs.get("choices", None):
            self.example = random.choice(kwargs["choices"])[0]
        else:
            self.example = "XYZ"
        super().__init__(*args, **kwargs)


class MoneyField(ModelExampleMixin, djmoney_fields.MoneyField):
    def add_currency_field(self, cls, name):
        """
        Adds CurrencyField instance to a model class and creates example in documentation.
        """

        currency_field = CurrencyField(
            price_field=self,
            max_length=self.currency_max_length,
            default=self.default_currency,
            choices=self.currency_choices,
            null=self.default_currency is None,
        )
        currency_field.creation_counter = self.creation_counter - 1
        currency_field_name = get_currency_field_name(name, self)
        cls.add_to_class(currency_field_name, currency_field)
        self._currency_field = currency_field

    def formfield(self, **kwargs):
        defaults = {
            "form_class": forms.MoneyField,
            "decimal_places": self.decimal_places,
        }
        defaults.update(kwargs)
        if self._has_default:
            defaults["default_amount"] = self.default.amount
        return super(djmoney_fields.MoneyField, self).formfield(**defaults)


class PhoneNumberField(ModelExampleMixin, PhoneNumberField):
    def __init__(self, *args, region=None, **kwargs) -> None:
        super().__init__(*args, region=region, **kwargs)
        if not kwargs.get("example", None):
            number = phonenumbers.example_number(region)
            self.example = str(to_python(number))


class CharField(ModelExampleMixin, models.CharField):
    def __init__(self, *args, **kwargs) -> None:
        if not kwargs.get("example") and kwargs.get("default"):
            kwargs["example"] = str(kwargs["default"])
        super().__init__(*args, **kwargs)
        max_length = kwargs.get("max_length", 80)
        if not kwargs.get("example", None) and max_length:
            self.example = generate_lorem_ipsum(max_length=max_length)


class TextField(ModelExampleMixin, models.TextField):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if not kwargs.get("example", None):
            self.example = generate_lorem_ipsum()


class MACAddressField(ModelExampleMixin, MACAddressField):
    pass


class JSONField(ModelExampleMixin, JSONField):
    pass


if "JSONField" not in __all__:
    __all__.append("JSONField")



import re
import typing
from copy import deepcopy
from inspect import isclass

from drf_spectacular.drainage import get_override
from drf_spectacular.extensions import OpenApiSerializerExtension
from drf_spectacular.openapi import AutoSchema
from drf_spectacular.plumbing import (
    ComponentRegistry,
    build_array_type,
    error,
    force_instance,
    sanitize_specification_extensions,
)
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiResponse
from rest_framework.fields import Field
from rest_framework.generics import GenericAPIView
from rest_framework.permissions import (
    AND,
    OR,
    BasePermission,
    OperandHolder,
    SingleOperandHolder,
)
from rest_framework.relations import RelatedField
from rest_framework.serializers import BaseSerializer
from rest_framework.views import APIView

from django.views import View

from audoma.decorators import SerializersConfig
from audoma.drf.generics import GenericAPIView as AudomaGenericAPIView
from audoma.drf.serializers import BulkSerializerMixin
from audoma.drf.validators import ExclusiveFieldsValidator
from audoma.links import (
    ChoicesOptionsLink,
    ChoicesOptionsLinkSchemaGenerator,
)
from audoma.plumbing import create_choices_enum_description


class AudomaAutoSchema(AutoSchema):
    choice_link_schema_generator = ChoicesOptionsLinkSchemaGenerator()

    def _handle_permission(
        self,
        permission_class: typing.Union[
            OperandHolder, SingleOperandHolder, BasePermission
        ],
        operations: list,
        current_operation: typing.Type = AND,
    ) -> dict:
        permissions = {}

        if isinstance(permission_class, OperandHolder):
            if permission_class.operator_class == OR and current_operation != OR:
                operations.append("(")
            permissions.update(
                self._handle_permission(
                    permission_class.op1_class,
                    operations,
                    permission_class.operator_class,
                )
            )
            if permission_class.operator_class == OR:
                operations.append("|")
            elif permission_class.operator_class == AND:
                operations.append(" & ")
            permissions.update(
                self._handle_permission(
                    permission_class.op2_class,
                    operations,
                    permission_class.operator_class,
                )
            )
            if permission_class.operator_class == OR and current_operation != OR:
                operations.append(" )")
        elif isinstance(permission_class, SingleOperandHolder):
            permissions.update(
                self._handle_permission(
                    permission_class.op1_class,
                    operations,
                    permission_class.operator_class,
                )
            )

        else:
            try:
                permissions[permission_class.__name__] = (
                    permission_class.get_description(self.view),
                )
            except AttributeError:
                if permission_class.__doc__:
                    permissions[
                        permission_class.__name__
                    ] = permission_class.__doc__.replace("\n", " ").strip()
                else:
                    permissions[
                        permission_class.__name__
                    ] = "(No description for this permission)"
            operations.append(f"`{permission_class.__name__}`")

        return permissions

    def _get_permissions_description(self) -> str:
        permissions = {}
        operations = []

        for permission_class in getattr(self.view, "permission_classes", []):
            if operations:
                operations.append("&")
            permissions.update(self._handle_permission(permission_class, operations))

        if permissions:
            return (
                "\n\n**Permissions:**\n"
                + " ".join(operations)
                + "\n"
                + "\n".join(
                    f"+ `{name}`: *{doc_str}*" for name, doc_str in permissions.items()
                )
            )
        else:
            return ""

    def get_description(self) -> str:
        description = super().get_description() or ""
        description += self._get_permissions_description()
        return description

    def _extract_action_function(self, view) -> typing.Callable:
        action = getattr(view, "action", None)
        if not action:
            return

        return getattr(view, action, None)

    def _parse_action_result_serializer(
        self,
        serializer: typing.Union[typing.Type[BaseSerializer], str],
        many: bool = False,
    ) -> typing.Union[BaseSerializer, str]:
        if isinstance(serializer, str):
            return OpenApiResponse(description=serializer)
        elif isclass(serializer) and issubclass(serializer, BaseSerializer):
            return serializer(many=many)
        return serializer

    def _parse_action_result_serializers(
        self, action_serializers: SerializersConfig, many: bool = False
    ) -> typing.Union[
        typing.Dict[str, OpenApiResponse],
        typing.Dict[str, BaseSerializer],
        typing.Dict[str, typing.Dict[int, BaseSerializer]],
    ]:
        if not action_serializers:
            return action_serializers

        if isinstance(action_serializers, str):
            return {"default": OpenApiResponse(description=action_serializers)}

        if not isinstance(action_serializers, dict):
            action_serializer = self._parse_action_result_serializer(
                action_serializers, many=many
            )
            return {"default": action_serializer}

        parsed_action_serializers = deepcopy(action_serializers)

        for method, method_serializers in action_serializers.items():
            if isinstance(method_serializers, dict):
                for code, item in method_serializers.items():
                    parsed_action_serializers[method][
                        code
                    ] = self._parse_action_result_serializer(item, many=many)
            else:
                parsed_action_serializers[
                    method
                ] = self._parse_action_result_serializer(method_serializers, many=many)

        return parsed_action_serializers

    def _parse_action_errors(self, action_errors) -> typing.Dict[int, OpenApiResponse]:
        if not action_errors:
            return action_errors

        parsed_errors = {}
        for err in action_errors:
            if isclass(err):
                err = err()

            # build properties
            properties = {}
            for key, value in vars(err).items():
                properties[key] = {key: {"type": type(value).__name__}}

            parsed_errors[err.status_code] = OpenApiResponse(
                response={
                    "type": "object",
                    "properties": properties,
                    "example": vars(err),
                }
            )
        return parsed_errors

    def _extract_audoma_action_operations(
        self, view: View, serializer_type: str
    ) -> typing.Union[
        typing.Dict[str, OpenApiResponse],
        typing.Dict[str, BaseSerializer],
        typing.Dict[str, typing.Type[BaseSerializer]],
        typing.Dict[str, typing.Dict[int, BaseSerializer]],
        typing.Dict[int, OpenApiResponse],
    ]:
        """
        Extracts the audoma action operations from the view
        """
        _audoma = view.get_audoma_action_config()
        if _audoma is None:
            return None
        many = _audoma.many

        if serializer_type == "collect":
            action_serializers = getattr(_audoma, "collectors", None)
        else:
            results = self._parse_action_result_serializers(
                getattr(_audoma, "results", None), many
            )
            errors = self._parse_action_errors(getattr(_audoma, "errors", {}))
            if results:
                action_serializers = results
                action_serializers.update(errors)
            else:
                action_serializers = errors

        method = view.request.method.lower()
        if isinstance(action_serializers, dict) and method in action_serializers:
            action_serializers = action_serializers[method]

        return action_serializers

    def _get_serializer(  # noqa: C901
        self, serializer_type: str = "collect"
    ) -> typing.Union[BaseSerializer, typing.Type[BaseSerializer]]:
        view = self.view
        try:
            if isinstance(view, AudomaGenericAPIView):
                action_serializers = self._extract_audoma_action_operations(
                    view, serializer_type
                )
                if action_serializers:
                    return action_serializers
                return view.get_serializer_class(serializer_type=serializer_type)()
            elif isinstance(view, GenericAPIView):
                # try to circumvent queryset issues with calling get_serializer. if view has NOT
                # overridden get_serializer, its safe to use get_serializer_class.
                if view.__class__.get_serializer == GenericAPIView.get_serializer:
                    return view.get_serializer_class()()
                return view.get_serializer()
            elif isinstance(view, APIView):
                # APIView does not implement the required interface, but be lenient and make
                # good guesses before giving up and emitting a warning.
                if callable(getattr(view, "get_serializer", None)):
                    return view.get_serializer()
                elif callable(getattr(view, "get_serializer_class", None)):
                    return view.get_serializer_class()()
                elif hasattr(view, "serializer_class"):
                    return view.serializer_class
                else:
                    error(
                        "unable to guess serializer. This is graceful "
                        "fallback handling for APIViews. Consider using GenericAPIView as view base "
                        "class, if view is under your control. Ignoring view for now. "
                    )
            else:
                error(
                    "Encountered unknown view base class. Please report this issue. Ignoring for now"
                )
        except Exception as exc:
            error(
                f"exception raised while getting serializer. Hint: "
                f"Is get_serializer_class() returning None or is get_queryset() not working without "
                f"a request? Ignoring the view for now. (Exception: {exc})"
            )

    def get_response_serializers(
        self,
    ) -> typing.Union[BaseSerializer, typing.Type[BaseSerializer]]:
        """overrides this for custom behaviour"""
        return self._get_serializer(serializer_type="result")

    def _get_enum_choices_for_field(self, field):
        if hasattr(field, "original_choices"):
            choices = field.original_choices
        else:
            choices = field.choices

        if hasattr(choices, "items"):
            choices = choices.items()

        return {"choices": {key: value for key, value in choices}}

    def _get_link_choices_for_field(self, field, serializer):
        link = serializer.choices_options_links.get(field.field_name, None)
        if not link:
            return

        if isinstance(link, dict):
            # presume that this dictionary are link kwargs
            link.update(
                {"field_name": field.field_name, "serializer_class": type(serializer)}
            )
            link = ChoicesOptionsLink(**link)

        choices = self.choice_link_schema_generator.generate_schema(link)
        if choices:
            if link.field_name == field.field_name:
                return choices
        return

    def _map_serializer_field(
        self, field: Field, direction: str, bypass_extensions=False
    ) -> dict:
        """
        Allows to use @extend_schema_field with `field` dict so that
        it gets updated instead of being overriden
        """
        serializer_type = "collect" if direction == "request" else "result"
        serializer = self._get_serializer(serializer_type=serializer_type)
        serializer = force_instance(serializer)

        if (
            hasattr(serializer, "Meta")
            and self.is_bulk
            and field.field_name == getattr(serializer.Meta, "id_field", None)
        ):
            field.read_only = False
            field.required = True

        has_annotation = (
            hasattr(field, "_spectacular_annotation")
            and "field" in field._spectacular_annotation
            and isinstance(field._spectacular_annotation["field"], dict)
        )
        if has_annotation:
            annotation = field._spectacular_annotation
            field._spectacular_annotation = {}

        result = super()._map_serializer_field(
            field, direction, bypass_extensions=bypass_extensions
        )

        if not isinstance(field, RelatedField) and hasattr(field, "choices"):
            result["x-choices"] = self._get_enum_choices_for_field(field)
            result["description"] = create_choices_enum_description(
                result["x-choices"]["choices"], field.field_name
            )

        if hasattr(serializer, "choices_options_links"):
            choices = self._get_link_choices_for_field(field, serializer)
            if choices:
                result["x-choices"] = choices

        if has_annotation:
            result.update(annotation["field"])

        return result

    def _get_request_for_media_type(self, serializer):

        schema, request_body_required = super()._get_request_for_media_type(serializer)
        if self.is_bulk:
            schema = build_array_type(schema)
        if isinstance(serializer, BulkSerializerMixin) and self.view.action == "create":
            schema = {"oneOf": [build_array_type(schema), schema]}
        return schema, request_body_required

    def _get_response_for_code(self, serializer, status_code, media_types=None):

        schema_resp = super()._get_response_for_code(
            serializer, status_code, media_types
        )

        if self.is_bulk:
            for media_type in schema_resp["content"]:
                schema = schema_resp["content"][media_type]["schema"]
                schema_resp["content"][media_type]["schema"] = build_array_type(schema)

        if isinstance(serializer, BulkSerializerMixin) and self.view.action == "create":
            for media_type in schema_resp["content"]:
                schema = schema_resp["content"][media_type]["schema"]
                schema_resp["content"][media_type]["schema"] = {
                    "oneOf": [build_array_type(schema), schema]
                }
        return schema_resp

    def _build_exclusive_fields_schema(
        self, schema: dict, exclusive_fields: typing.List[str]
    ) -> typing.List[dict]:
        modified_schemas = []
        for field in exclusive_fields:
            new_schema = deepcopy(schema)
            new_schema["properties"].pop(field)
            modified_schemas.append(new_schema)
        return modified_schemas

    def _map_serializer(
        self,
        serializer: typing.Union[
            OpenApiTypes, BaseSerializer, typing.Type[BaseSerializer]
        ],
        direction: str,
        bypass_extensions: bool = False,
    ) -> dict:
        serializer = force_instance(serializer)
        serializer_extension = OpenApiSerializerExtension.get_match(serializer)

        if serializer_extension and not bypass_extensions:
            schema = serializer_extension.map_serializer(self, direction)
        else:
            schema = self._map_basic_serializer(serializer, direction)

        if hasattr(serializer, "validators") and direction == "request":
            subschemas = []
            for validator in serializer.validators:
                if isinstance(validator, ExclusiveFieldsValidator):
                    subschemas += self._build_exclusive_fields_schema(
                        schema, validator.fields
                    )

            if subschemas:
                schema = {"oneOf": subschemas}

        extensions = get_override(serializer, "extensions", {})
        if extensions:
            schema.update(sanitize_specification_extensions(extensions))

        return self._postprocess_serializer_schema(schema, serializer, direction)

    def get_operation_id(self):
        """override this for custom behaviour"""
        tokenized_path = self._tokenize_path()

        if self.is_bulk:
            tokenized_path.append("bulk")

        # replace dashes as they can be problematic later in code generation
        tokenized_path = [t.replace("-", "_") for t in tokenized_path]

        if self.method == "GET" and self._is_list_view():
            action = "list"
        else:
            action = self.method_mapping[self.method.lower()]

        if not tokenized_path:
            tokenized_path.append("root")

        if re.search(r"<drf_format_suffix\w*:\w+>", self.path_regex):
            tokenized_path.append("formatted")

        return "_".join(tokenized_path + [action])

    def get_operation(
        self, path, path_regex, path_prefix, method, registry: ComponentRegistry
    ):
        self.is_bulk = False

        if "bulk" in getattr(self.view, "action", "") and method in [
            "POST",
            "PUT",
            "PATCH",
        ]:
            self.is_bulk = True

        return super().get_operation(path, path_regex, path_prefix, method, registry)



import random
from typing import (
    Any,
    Dict,
    List,
    Union,
)

from rest_framework import viewsets
from rest_framework.exceptions import ErrorDetail
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response

from django.conf import settings

from audoma.drf.generics import GenericAPIView


class AudomaPagination(PageNumberPagination):
    """
    A simple page number based style that supports page numbers as
    query parameters.

    Note:
        If this won't be used it'll cause less explicit pagination documentation.
        This class does not provide any additional functionality.

    Args:
        page_size (int) - number of items per page - by default this is set to 25
        max_page_size (int) - maximum number of items per page - by default this is set to 2000
    """

    page_size = 25
    max_page_size = 2000

    def get_paginated_response_schema(self, schema: List[dict]) -> dict:
        """
        Simple method to add pagination information to the schema.

        Args:
            schema (List[dict]) - list of schema items
        Returns:
            Dictionary with pagination information including examples
        """
        return {
            "type": "object",
            "properties": {
                "count": {"type": "integer", "example": random.randint(1, 100)},
                "message": {
                    "type": "string",
                    "nullable": True,
                },
                "next": {
                    "type": "string",
                    "nullable": True,
                    "format": "uri",
                    "example": "http://api.example.org/accounts/?page=4",
                },
                "previous": {
                    "type": "string",
                    "nullable": True,
                    "format": "uri",
                    "example": "http://api.example.org/accounts/?page=2",
                },
                "results": schema,
            },
        }


class UnknownExceptionContentTypeError(Exception):
    ...


class GenericViewSet(viewsets.ViewSetMixin, GenericAPIView):
    pagination_class = AudomaPagination

    def _parse_single_response_data_item(self, item):
        if isinstance(item, ErrorDetail):
            return str(item)
        elif isinstance(item, str):
            return item
        elif isinstance(item, list) or isinstance(item, dict):
            return self._parse_response_data(item)
        else:
            raise UnknownExceptionContentTypeError

    def _simplify_validation_errors(self, parsed_data):
        if (
            isinstance(parsed_data, list)
            and len(parsed_data) == 1
            and all(isinstance(d, str) for d in parsed_data)
        ):
            parsed_data = "".join(parsed_data)
        return parsed_data

    def _parse_response_data(
        self, response_data: Union[List[Any], Dict[Any, Any]]
    ) -> Union[List[Any], Dict[Any, Any]]:
        parsed_data = []
        if isinstance(response_data, list):
            for item in response_data:
                parsed_data.append(self._parse_single_response_data_item(item))
        elif isinstance(response_data, dict):
            for key, item in response_data.items():
                item = self._parse_single_response_data_item(item)
                parsed_data.append((key, item))
            parsed_data = dict(parsed_data)

        if getattr(settings, "AUDOMA_SIMPLIFY_VALIDATION_ERRORS", False):
            parsed_data = self._simplify_validation_errors(parsed_data)
        return parsed_data

    def handle_exception(self, exc: Exception) -> Response:
        response = super().handle_exception(exc)
        if isinstance(response.data, dict):
            if response.status_code != 418:
                for k in response.data:
                    if isinstance(response.data[k], list):
                        if not all(isinstance(item, str) for item in response.data[k]):
                            response.data[k] = self._parse_response_data(
                                response.data[k]
                            )
                        elif getattr(
                            settings, "AUDOMA_SIMPLIFY_VALIDATION_ERRORS", False
                        ):
                            response.data[k] = self._simplify_validation_errors(
                                response.data[k]
                            )

                response.data = {"errors": response.data}
        elif isinstance(response.data, list):
            response.data = {"errors": response.data}
        return response



from unittest import TestCase

from rest_framework import (
    generics,
    mixins,
    serializers,
    views,
    viewsets,
)

from audoma.drf import (
    generics as audoma_generics,
    mixins as audoma_mixins,
    serializers as audoma_serializers,
    viewsets as audoma_viewsets,
)
from audoma.plumbing import (
    create_choices_enum_description,
    get_lib_doc_excludes_audoma,
)


class PlumbingTestCase(TestCase):
    def setUp(self):
        super().setUp()

    def test_create_choices_enum_description_dict_choices(self):
        description = create_choices_enum_description(
            {1: "One", 2: "Two", 3: "Three"}, "example_field"
        )
        self.assertEqual(
            description,
            "Filter by example_field \n * `1` - One\n * `2` - Two\n * `3` - Three\n",
        )

    def test_create_choices_enum_description_non_dict_choices(self):
        description = create_choices_enum_description(
            [(1, "One"), (2, "Two"), (3, "Three")], "example_field"
        )
        self.assertEqual(
            description,
            "Filter by example_field \n * `1` - One\n * `2` - Two\n * `3` - Three\n",
        )

    def test_get_lib_doc_excludes_audoma(self):
        expected_exclude_list = [
            views.APIView,
            serializers.BaseSerializer,
            serializers.HyperlinkedModelSerializer,
            serializers.ListSerializer,
            serializers.ModelSerializer,
            serializers.Serializer,
            viewsets.GenericViewSet,
            viewsets.ModelViewSet,
            viewsets.ReadOnlyModelViewSet,
            viewsets.ViewSet,
            generics.CreateAPIView,
            generics.DestroyAPIView,
            generics.GenericAPIView,
            generics.ListAPIView,
            generics.ListCreateAPIView,
            generics.RetrieveAPIView,
            generics.RetrieveDestroyAPIView,
            generics.RetrieveUpdateAPIView,
            generics.RetrieveUpdateDestroyAPIView,
            generics.UpdateAPIView,
            mixins.CreateModelMixin,
            mixins.DestroyModelMixin,
            mixins.ListModelMixin,
            mixins.RetrieveModelMixin,
            mixins.UpdateModelMixin,
            audoma_viewsets.GenericViewSet,
            audoma_mixins.ActionModelMixin,
            audoma_mixins.BulkCreateModelMixin,
            audoma_mixins.BulkUpdateModelMixin,
            audoma_mixins.CreateModelMixin,
            audoma_mixins.DestroyModelMixin,
            audoma_mixins.ListModelMixin,
            audoma_mixins.RetrieveModelMixin,
            audoma_mixins.UpdateModelMixin,
            audoma_serializers.BaseSerializer,
            audoma_serializers.BulkListSerializer,
            audoma_serializers.DefaultMessageSerializer,
            audoma_serializers.HyperlinkedModelSerializer,
            audoma_serializers.ListSerializer,
            audoma_serializers.ModelSerializer,
            audoma_serializers.Serializer,
            audoma_generics.GenericAPIView,
        ]
        exclude_list = get_lib_doc_excludes_audoma()
        self.assertEqual(exclude_list, expected_exclude_list)



"""
This module is responsible for creating x-choices links.
Such links may be used to generate choices enums for fields.

If some field in the serializer has this attribute, it
will be used to generate choices for this field.
In other words choices should be limited to values
available under passed x-choices link.
"""

import re
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    Type,
    Union,
)

from drf_spectacular.plumbing import force_instance
from rest_framework.serializers import BaseSerializer

from django.urls import (
    NoReverseMatch,
    URLResolver,
)
from django.urls.resolvers import get_resolver


def get_endpoint_pattern(endpoint_name: str, urlconf=None) -> str:
    """
    This methods retrieves url pattern of the endpoint by given endpoint_name.

    Args:
        * endpoint_name: name of the endpoint
        * urlconf: urlconf to use

    Returns: url pattern of the endpoint
    """
    resolver = get_resolver(urlconf)
    patterns = resolver.url_patterns
    new_patterns = []
    resolvers = []

    for pattern in patterns:
        if isinstance(pattern, URLResolver):
            resolvers.append(pattern)
            continue
        new_patterns.append(pattern)

    while resolvers:
        resolver = resolvers.pop()
        patterns = resolver.url_patterns
        for pattern in patterns:
            if isinstance(pattern, URLResolver):
                resolvers.append(pattern)
                continue
            new_patterns.append(pattern)

    possibilities = list(filter(lambda p: p.name == endpoint_name, new_patterns))

    for p in possibilities:
        if "format" not in p.pattern.regex.pattern:
            return str(p.pattern)

    raise NoReverseMatch(f"There is no pattern with name {endpoint_name}")


@dataclass
class ChoicesOptionsLink:
    """
    Helper dataclass, which holds data abot x-choices link.
    """

    field_name: str
    viewname: str
    value_field: str
    display_field: str
    serializer_class: Type[BaseSerializer]

    description: str = ""

    def _format_param_field(self, field: str) -> str:
        """
        Helper method which formats field name to be a JSON pointer.
        If the passed field name is already a JSON pointer, it is returned unchagned.

        Args:
            * field: field name

        Returns: formatted field name
        """
        if "$" in field:
            # than we presume that there has been given full pointer
            return field

        field = f"$response.body#results/*/{field}"

        return field

    @property
    def formatted_value_field(self) -> str:
        return self._format_param_field(self.value_field)

    @property
    def formatted_display_field(self) -> str:
        return self._format_param_field(self.display_field)

    def get_url_pattern(self) -> str:
        """
        Returns formatted url pattern of the linked view.
        """
        pattern = (
            get_endpoint_pattern(self.viewname)
            .replace("$", "")
            .replace("^", "/")
            .replace("/", "~1")
        )

        params = re.search(r"<.*>", pattern)
        if params:
            for param in params:
                param = param.string.replace("<", "{").replace(">", "}")

            regexes = re.search(r"\(.*\)", pattern)
            for x, regex in enumerate(regexes):
                pattern = pattern.replace(regex.string, params[x])

        pattern = f"#/paths/{pattern}"

        return pattern


class ChoicesOptionsLinkSchemaGenerator:
    def _process_link(
        self, link: Union[ChoicesOptionsLink, Dict[str, Any]]
    ) -> ChoicesOptionsLink:
        if isinstance(link, dict):
            link = ChoicesOptionsLink(**link)

        if not isinstance(link, ChoicesOptionsLink):
            raise TypeError(
                f"This is not possible to create ChoicesOptionsLink \
                    from object of type {type(link)}"
            )

        serializer = force_instance(link.serializer_class)
        # serializer must own defined field
        if not serializer.fields.get(link.field_name, None):
            raise AttributeError(
                f"Serializer class: {link.serializer_class} does not have field: {link.field_name}"
            )
        return link

    def _create_link_title(self, link: Union[ChoicesOptionsLink, dict]) -> str:
        partials = link.viewname.replace("-", " ").replace("_", " ").split(" ")
        partials = [p.capitalize() for p in partials]
        return " ".join(partials).title()

    def generate_schema(self, link: Union[ChoicesOptionsLink, Dict[str, Any]]) -> dict:
        """
        Generates x-choices link schema.
        Args:
            * link: ChoicesOptionsLink instance or dict with link data

        Returns: x-choices link schema
        """

        if not link:
            return

        link = self._process_link(link)

        schema = {
            "operationRef": link.get_url_pattern(),
            # "parameters": "",
            "value": link.formatted_value_field,
            "display": link.formatted_display_field,
        }
        return schema



import json
import re
from datetime import (
    date,
    timedelta,
)
from typing import OrderedDict

import phonenumbers
from audoma_api.models import (
    Car,
    ExampleModel,
    Manufacturer,
)
from audoma_api.serializers import (
    ExampleModelSerializer,
    ExampleSerializer,
)
from audoma_api.views import (
    ExampleModelPermissionLessViewSet,
    ExampleModelViewSet,
    ExampleViewSet,
)
from drf_example.urls import router
from drf_spectacular.generators import SchemaGenerator
from phonenumber_field.phonenumber import to_python
from rest_framework.exceptions import ErrorDetail
from rest_framework.permissions import BasePermission
from rest_framework.test import (
    APIClient,
    APIRequestFactory,
    APITestCase,
)

from django.conf import settings
from django.shortcuts import reverse
from django.test import (
    SimpleTestCase,
    TestCase,
    override_settings,
)

from audoma.decorators import AudomaActionException
from audoma.django.db import models
from audoma.drf import serializers
from audoma.drf.viewsets import AudomaPagination
from audoma.example_generators import generate_lorem_ipsum


class AudomaTests(SimpleTestCase):
    def setUp(self):
        patterns = router.urls
        generator = SchemaGenerator(patterns=patterns)
        self.schema = generator.get_schema(request=None, public=True)
        self.redoc_schemas = self.schema["components"]["schemas"]

    def test_extend_schema_field_with_example_not_override_format(self):
        phone_number_field = self.redoc_schemas["Example"]["properties"][
            "phone_number_example"
        ]
        expected_phone_number_field = {
            "type": "string",
            "format": "tel",
            "example": "+48 123 456 789",
        }
        self.assertEqual(expected_phone_number_field, phone_number_field)

    def test_extend_schema_field_with_example_as_init(self):
        date_field = self.redoc_schemas["Example"]["properties"]["date"]
        time_field = self.redoc_schemas["Example"]["properties"]["time"]
        expected_date_field = {
            "type": "string",
            "format": "date",
            "example": str(date.today()),
        }
        expected_time_field = {
            "type": "string",
            "format": "time",
            "example": "12:34:56.000000",
        }
        self.assertEqual(expected_date_field, date_field)
        self.assertEqual(expected_time_field, time_field)

    def test_extend_schema_field_with_openapitype(self):
        uuid_field = self.redoc_schemas["Example"]["properties"]["uuid"]
        expected_uuid_field = {"type": "string", "format": "uuid"}
        self.assertEqual(expected_uuid_field, uuid_field)

    def test_model_mapping_all_field_serializer(self):
        example_model_properties = self.redoc_schemas["ExampleModel"]["properties"]
        self.assertEqual(len(ExampleModel._meta.fields), len(example_model_properties))

    def test_permission_description_extension_model_viewset(self):
        expected_permissions = ExampleModelViewSet.permission_classes
        description = self.schema["paths"]["/model_examples/"]["get"]["description"]
        for permission in expected_permissions:
            # skip operations
            if not isinstance(permission, BasePermission):
                continue
            self.assertIn(str(permission.__name__), description)

    def test_permission_description_extension_standard_viewset(self):
        expected_permissions = ExampleViewSet.permission_classes
        description = self.schema["paths"]["/examples/"]["get"]["description"]
        for permission in expected_permissions:
            # skip operations
            if not isinstance(permission, BasePermission):
                continue
            self.assertIn(str(permission.__name__), description)

    def test_document_and_format_example_model_phone_number(self):
        example_model_properties = self.redoc_schemas["ExampleModel"]["properties"]
        phone_number = example_model_properties["phone_number"]
        self.assertEqual("tel", phone_number["format"])

    def test_custom_paginated_response_schema(self):
        paginated_example = self.redoc_schemas["PaginatedExampleList"]
        paginator = AudomaPagination()
        expected_pagination = paginator.get_paginated_response_schema(paginated_example)
        self.assertEqual(
            paginated_example["properties"].keys(),
            expected_pagination["properties"].keys(),
        )

    def test_example_float_field_with_range(self):
        float_field = self.redoc_schemas["Example"]["properties"]["float"]
        self.assertLessEqual(float_field["minimum"], float_field["example"])
        self.assertGreaterEqual(float_field["maximum"], float_field["example"])

    def test_example_mac_address_field_with_regex_mixin(self):
        example_mac_address = self.redoc_schemas["Example"]["properties"]["mac_address"]
        mac_address = serializers.MACAddressField()
        regex_pattern = re.compile(mac_address.regex)
        self.assertEqual(mac_address.regex, example_mac_address["pattern"])
        self.assertTrue(bool(regex_pattern.match(example_mac_address["example"])))

    def test_override_model_example_in_extra_kwargs(self):
        example_model_properties = self.redoc_schemas["ExampleModel"]["properties"]
        char_field = example_model_properties["char_field"]
        expected_result = ExampleModelSerializer.Meta.extra_kwargs["char_field"][
            "example"
        ]
        self.assertEqual(expected_result, char_field["example"])

    def test_example_models_custom_examples(self):
        example_person_properties = self.redoc_schemas["ExamplePersonModel"][
            "properties"
        ]
        first_name = example_person_properties["first_name"]
        last_name = example_person_properties["last_name"]
        self.assertEqual("Adam", first_name["example"])
        self.assertEqual("Smith", last_name["example"])

    def test_example_with_callable_as_argument(self):
        example_person_properties = self.redoc_schemas["ExamplePersonModel"][
            "properties"
        ]
        age = example_person_properties["age"]
        self.assertLessEqual(18, age["example"])
        self.assertGreaterEqual(80, age["example"])

    def test_model_phonenumber_with_region(self):
        example_person_properties = self.redoc_schemas["ExamplePersonModel"][
            "properties"
        ]
        phone_number_field = example_person_properties["phone_number"]
        phone_number_example = phone_number_field["example"]
        generated_france_number = to_python(
            phonenumbers.example_number("IT")
        ).as_international
        self.assertEqual(generated_france_number, phone_number_example)

    def test_phonenubber_example_region_international_format(self):

        pn = models.PhoneNumberField(region="PL")

        expected_example = to_python(phonenumbers.example_number("PL")).as_international
        self.assertEqual(expected_example, pn.example)

    @override_settings(PHONENUMBER_DEFAULT_FORMAT="NATIONAL")
    def test_phonenumber_example_region_national_format(self):
        pn = models.PhoneNumberField(region="PL")

        expected_example = to_python(phonenumbers.example_number("PL")).as_national
        self.assertEqual(expected_example, pn.example)

    @override_settings(PHONENUMBER_DEFAULT_FORMAT="E164")
    def test_phonenumber_example_region_e164_format(self):

        pn = models.PhoneNumberField(region="PL")

        expected_example = to_python(phonenumbers.example_number("PL")).as_e164
        self.assertEqual(expected_example, pn.example)

    def test_file_upload_view_parsers(self):
        example_schema = self.schema["paths"]["/file-upload-example/"]["post"][
            "requestBody"
        ]["content"]
        self.assertEqual(len(example_schema.keys()), 1)
        self.assertEqual(list(example_schema.keys())[0], "multipart/form-data")

    def test_filterset_class_description_in_query_params_schema(self):
        choices_desc = "Filter by choices \n * `EX_1` - example 1\n * `EX_2` - example 2\n * `EX_3` - example 3\n"
        docs_description = self.schema["paths"]["/model_examples/"]["get"][
            "parameters"
        ][0]["description"]
        self.assertEqual(choices_desc, docs_description)

    def test_filterset_fields_description_in_query_paramas_schema(self):
        choices_desc = "Filter by engine_type \n * `1` - Petrol\n * `2` - Diesel\n * `3` - Electric\n * `4` - Hybrid\n"
        docs_description = self.schema["paths"]["/car_viewset/"]["get"]["parameters"][
            0
        ]["description"]
        self.assertEqual(choices_desc, docs_description)

    def test_search_fields_description(self):
        expected_search_description = (
            "Search by: \n* `manufacturer(Exact matches.)` \n* `name` \n"
        )

        docs_description = self.schema["paths"]["/car_viewset/"]["get"]["parameters"]
        search_docs_data = docs_description[-1]
        self.assertEqual(search_docs_data["name"], "search")
        self.assertEqual(search_docs_data["description"], expected_search_description)

    def test_x_choices_enum_serializer(self):
        expected_choices = {
            1: "Sedan",
            2: "Coupe",
            3: "Hatchback",
            4: "Pickup Truck",
        }
        choices_schema = self.schema["components"]["schemas"]["CarModel"]["properties"][
            "body_type"
        ]["x-choices"]["choices"]
        for key, item in choices_schema.items():
            self.assertEqual(item, expected_choices[key])

    def test_x_choices_enum_paramteres(self):
        expected_choices = {
            1: "Petrol",
            2: "Diesel",
            3: "Electric",
            4: "Hybrid",
        }
        choices_schema = self.schema["paths"]["/car_viewset/"]["get"]["parameters"][0][
            "schema"
        ]["x-choices"]["choices"]

        for key, item in choices_schema.items():
            self.assertEqual(item, expected_choices[key])

    def test_x_choices_link_serializer(self):
        expected_link = {
            "operationRef": "#/paths/~1manufacturer_viewset~1",
            "value": "$response.body#results/*/id",
            "display": "$response.body#results/*/name",
        }
        link_schema = self.schema["components"]["schemas"]["CarModel"]["properties"][
            "manufacturer"
        ]["x-choices"]
        for key, item in link_schema.items():
            self.assertEqual(item, expected_link[key])

    def test_charfield_example_limits(self):
        charfield_redoc = self.redoc_schemas["Example"]["properties"][
            "charfield_min_max"
        ]
        charfield_example = charfield_redoc["example"]
        charfield = ExampleSerializer.__dict__["_declared_fields"]["charfield_min_max"]
        min_length = charfield.min_length
        max_length = charfield.max_length
        self.assertLessEqual(min_length, len(charfield_example))
        self.assertGreaterEqual(max_length, len(charfield_example))

    def test_generate_lorem_ipsum_only_min_length(self):
        short_lorem_example = generate_lorem_ipsum(min_length=60)
        long_lorem_example = generate_lorem_ipsum(min_length=1024)
        self.assertGreaterEqual(len(short_lorem_example), 60)
        self.assertLessEqual(len(short_lorem_example), 80)
        self.assertGreaterEqual(len(long_lorem_example), 1024)

    def test_generate_lorem_ipsum_only_max_length(self):
        short_lorem_example = generate_lorem_ipsum(max_length=10)
        long_lorem_example = generate_lorem_ipsum(max_length=1024)
        self.assertLessEqual(len(short_lorem_example), 10)
        self.assertGreaterEqual(len(long_lorem_example), 20)
        self.assertLessEqual(len(long_lorem_example), 1024)

    def test_generate_lorem_ipsum_max_min(self):
        short_lorem_example = generate_lorem_ipsum(min_length=1, max_length=5)
        long_lorem_example = generate_lorem_ipsum(min_length=100, max_length=200)
        self.assertGreaterEqual(len(short_lorem_example), 1)
        self.assertLessEqual(len(short_lorem_example), 5)
        self.assertGreaterEqual(len(long_lorem_example), 100)
        self.assertLessEqual(len(long_lorem_example), 200)

    def test_example_money_currency_with_currency_from_settings(self):
        currency = self.redoc_schemas["ExampleModel"]["properties"]["money_currency"]
        self.assertIn(currency["example"], settings.CURRENCIES)

    def test_example_money_currency_with_default_currency(self):
        currency = self.redoc_schemas["ExamplePersonModel"]["properties"][
            "savings_currency"
        ]
        self.assertEqual("PLN", currency["example"])

    def test_mutually_exclusive_fields_schemas(self):
        schema = self.redoc_schemas["MutuallyExclusiveExampleRequest"]
        default_keys = ["not_exclusive_field", "second_not_exclusive_field"]
        self.assertEqual(len(schema["oneOf"]), 4)
        self.assertEqual(
            list(schema["oneOf"][0]["properties"].keys()),
            ["second_example_field", "third_example_field", "fourth_example_field"]
            + default_keys,
        )
        self.assertEqual(
            list(schema["oneOf"][1]["properties"].keys()),
            ["example_field", "third_example_field", "fourth_example_field"]
            + default_keys,
        )
        self.assertEqual(
            list(schema["oneOf"][2]["properties"].keys()),
            ["example_field", "second_example_field", "fourth_example_field"]
            + default_keys,
        )
        self.assertEqual(
            list(schema["oneOf"][3]["properties"].keys()),
            ["example_field", "second_example_field", "third_example_field"]
            + default_keys,
        )

    def test_default_response_in_view_responses(self):
        docs = self.schema["paths"][
            "/permissionless_model_examples/properly_defined_exception_example/"
        ]
        responses_docs = docs["get"]["responses"]
        self.assertEqual(
            responses_docs["default"]["content"]["application/json"]["schema"]["$ref"],
            "#/components/schemas/ExampleOneField",
        )

    def test_custom_responses_in_view_responses(self):
        docs = self.schema["paths"][
            "/permissionless_model_examples/{id}/detail_action/"
        ]
        responses_docs = docs["post"]["responses"]
        self.assertEqual(
            responses_docs["201"]["content"]["application/json"]["schema"]["$ref"],
            "#/components/schemas/ExampleModel",
        )
        self.assertEqual(
            responses_docs["202"]["content"]["application/json"]["schema"]["$ref"],
            "#/components/schemas/ExampleOneField",
        )

    def test_common_errors_in_description(self):
        expected_error_data = '\n {\n    "detail": "Not found."\n} \n'
        self.assertIn(expected_error_data, self.schema["info"]["description"])
        expected_error_data = ' \n {\n    "detail": "Authentication credentials were not provided."\n} \n '
        self.assertIn(expected_error_data, self.schema["info"]["description"])

    def test_viewset_errors_in_viewset_responses(self):
        docs = self.schema["paths"][
            "/permissionless_model_examples/properly_defined_exception_example/"
        ]
        responses_docs = docs["get"]["responses"]
        self.assertEqual(
            responses_docs["409"]["content"]["application/json"]["schema"]["example"][
                "detail"
            ],
            "Conflict has occured",
        )
        self.assertEqual(
            responses_docs["400"]["content"]["application/json"]["schema"]["example"][
                "detail"
            ],
            "Custom Bad Request Exception",
        )


class AudomaBulkOperationsTest(APITestCase):
    def setUp(self):
        self.list_url = reverse("bulk-example-list")
        Manufacturer.objects.bulk_create(
            [
                Manufacturer(name="Example 1", slug_name="ex_1"),
                Manufacturer(name="Example 2", slug_name="ex_2"),
            ]
        )
        return super().setUp()

    def test_bulk_create_records(self):
        data = [
            {
                "name": "test 1",
                "slug_name": "2138",
            },
            {
                "name": "test 2",
                "slug_name": "2137",
            },
        ]

        resp = self.client.post(self.list_url, data, format="json")
        self.assertEqual(201, resp.status_code, resp.json())

        qs = Manufacturer.objects.all()
        self.assertEqual(qs.count(), 4)

    def test_bulk_nested_errors(self):
        data = [
            {
                "name": "test 1",
            },
            {
                "name": "test 2",
            },
        ]

        resp = self.client.post(self.list_url, data, format="json")
        self.assertEqual(400, resp.status_code)
        content = json.loads(resp.content)
        self.assertEqual(
            content["errors"],
            [
                {"slug_name": ["This field is required."]},
                {"slug_name": ["This field is required."]},
            ],
        )

    def test_bulk_create_single_record(self):
        data = (
            {
                "name": "test 2",
                "slug_name": "2137",
            },
        )

        resp = self.client.post(self.list_url, data, format="json")
        self.assertEqual(201, resp.status_code, resp.json())

        qs = Manufacturer.objects.all()
        self.assertEqual(qs.count(), 3)

    def test_bulk_update_records(self):
        updated_data = [
            {
                "id": 1,
                "name": "test 1 updated",
                "slug_name": "11",
            },
            {
                "id": 2,
                "name": "test 2 updated",
                "slug_name": "22",
            },
        ]

        resp = self.client.put(self.list_url, updated_data, format="json")
        qs = Manufacturer.objects.all()
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(qs.first().name, "test 1 updated")
        self.assertEqual(qs.last().name, "test 2 updated")
        self.assertEqual(qs.first().slug_name, "11")
        self.assertEqual(qs.last().slug_name, "22")

    def test_bulk_partial_update_records(self):

        updated_data = [
            {
                "id": 1,
                "name": "test 1 updated",
            },
            {
                "id": 2,
                "name": "test 2 updated",
            },
        ]

        resp = self.client.patch(self.list_url, updated_data, format="json")
        qs = Manufacturer.objects.all()
        self.assertEqual(resp.status_code, 200, resp.json())
        self.assertEqual(qs.first().name, "test 1 updated")
        self.assertEqual(qs.last().name, "test 2 updated")
        self.assertEqual(qs.first().slug_name, "ex_1")
        self.assertEqual(qs.last().slug_name, "ex_2")

    # def test_bulk_delete_records(self):
    #     ExampleSimpleModel.objects.bulk_create(
    #         [
    #             ExampleSimpleModel(name="EX1", value=1),
    #             ExampleSimpleModel(name="EX2", value=2),
    #         ]
    #     )
    #     resp = self.client.delete(
    #         self.list_url, data=[{"name": "EX1"}, {"name": "EX2"}], format="json"
    #     )
    #     self.assertEqual(resp.status_code, 204)
    #     self.assertEqual(
    #         ExampleSimpleModel.objects.filter(name__in=["EX1", "EX2"]).count(), 0
    #     )

    def test_bulk_partial_update_records_invalid_pk(self):

        updated_data = [
            {
                "id": 4,
                "name": "test 1 updated",
            },
            {
                "id": 5,
                "name": "test 2 updated",
            },
        ]

        resp = self.client.patch(self.list_url, updated_data, format="json")
        self.assertEqual(resp.status_code, 400)

    # def test_bulk_delete_validation(self):
    #     ExampleSimpleModel.objects.bulk_create(
    #         [
    #             ExampleSimpleModel(name="EX1", value=1),
    #             ExampleSimpleModel(name="EX2", value=2),
    #         ]
    #     )
    #     resp = self.client.delete(
    #         self.list_url, data=[{"name": "EX2"}, {"name": "EX3"}], format="json"
    #     )
    #     self.assertEqual(resp.status_code, 400)
    #     self.assertEqual(
    #         ExampleSimpleModel.objects.filter(name__in=["EX1", "EX2"]).count(), 0
    #     )


class AudomaViewsTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.data = {
            "char_field": "TESTChar",
            "phone_number": "+18888888822",
            "phone_number_example": "+18888888822",
            "email": "test@iteo.com",
            "url": "http://localhost:8000/redoc/",
            "boolean": False,
            "nullboolean": None,
            "mac_adress": "96:82:2E:6B:F5:49",
            "slug": "tst",
            "uuid": "14aefe15-7c96-49b6-9637-7019c58c25d2",
            "ip_address": "192.168.10.1",
            "integer": 16,
            "_float": 12.2,
            "decimal": "13.23",
            "datetime": "2009-11-13T10:39:35Z",
            "date": "2009-11-13",
            "time": "10:39:35",
            "duration": timedelta(days=1),
            "choices": 1,
            "json": "",
            "money": "12.23",
            "text_field": "TESTText",
        }
        self.client = APIClient()

    def test_detail_action_get(self):
        response = self.client.get(
            reverse("permissionless-model-examples-detail-action", kwargs={"pk": 0})
        )
        self.assertEqual(response.status_code, 405)

    def test_detail_action_post_with_usertype(self):
        self.data["usertype"] = "admin"
        response = self.client.post(
            reverse("permissionless-model-examples-detail-action", kwargs={"pk": 0}),
            self.data,
            format="json",
        )
        self.assertEqual(response.status_code, 201)
        self.assertTrue(response.has_header("CustomHeader"))
        response_content = json.loads(response.content)
        self.assertEqual(self.data["char_field"], response_content["char_field"])
        self.assertEqual(self.data["mac_adress"], response_content["mac_adress"])
        self.assertEqual(self.data["uuid"], response_content["uuid"])

    def test_detail_action_post_without_usertype(self):
        response = self.client.post(
            reverse("permissionless-model-examples-detail-action", kwargs={"pk": 0}),
            self.data,
            format="json",
        )
        self.assertEqual(response.status_code, 202)
        self.assertTrue(response.has_header("CustomHeader"))
        response_content = json.loads(response.content)
        self.assertEqual(response_content["rate"], 0)

    def test_non_detail_action_get(self):
        response = self.client.get(
            reverse("permissionless-model-examples-non-detail-action")
        )
        content = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(content["message"], "This is a test view")

    def test_rate_create_action_post_success(self):
        response = self.client.post(
            reverse("permissionless-model-examples-rate-create-action"),
            {"rate": 1},
            format="json",
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(json.loads(response.content)["rate"], 1)

    def test_rate_create_action_post_failure(self):
        response = self.client.post(
            reverse("permissionless-model-examples-rate-create-action")
        )
        content = json.loads(response.content)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(content["errors"]["rate"], ["This field is required."])

    @override_settings(AUDOMA_SIMPLIFY_VALIDATION_ERRORS=True)
    def test_rate_create_action_post_failure_join_errors(self):
        response = self.client.post(
            reverse("permissionless-model-examples-rate-create-action")
        )
        content = json.loads(response.content)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(content["errors"]["rate"], "This field is required.")

    def test_specific_rate_get_success(self):
        response = self.client.get(
            reverse("permissionless-model-examples-specific-rate", kwargs={"pk": 1})
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["rate"], 1)

    @override_settings(DEBUG=True)
    def test_properly_defined_exception_example(self):
        response = self.client.get(
            reverse("permissionless-model-examples-properly-defined-exception-example")
        )
        self.assertEqual(response.status_code, 409)
        content = json.loads(response.content)
        self.assertEqual(
            content["errors"]["detail"], "Some custom message, that should be accepted"
        )

    def test_proper_usage_of_common_errors(self):
        response = self.client.get(
            reverse("permissionless-model-examples-proper-usage-of-common-errors")
        )
        self.assertEqual(response.status_code, 404)
        content = json.loads(response.content)
        self.assertEqual(content["errors"]["detail"], "Not found.")

    @override_settings(DEBUG=True)
    def test_improperly_defined_exception_example_debug_true(self):
        try:
            view = ExampleModelPermissionLessViewSet()
            url = reverse(
                "permissionless-model-examples-improperly-defined-exception-example"
            )
            factory = APIRequestFactory()
            request = factory.get(url)
            view.improperly_defined_exception_example(request)

        except Exception as e:
            self.assertEqual(type(e), AudomaActionException)
            self.assertIn(
                "<class 'audoma_api.exceptions.CustomBadRequestException'>", str(e)
            )

    def test_improperly_defined_exception_example_debug_false(self):
        response = self.client.get(
            reverse(
                "permissionless-model-examples-improperly-defined-exception-example"
            )
        )
        self.assertEqual(response.status_code, 400)
        data = response.json()
        self.assertEqual(data["errors"]["detail"], "Custom Bad Request Exception")

    def test_example_update_action_patch(self):
        response = self.client.patch(
            reverse(
                "permissionless-model-examples-example-update-action", kwargs={"pk": 0}
            ),
            {"char_field": "TESTChar2", "email": "changetest@iteo.com"},
            format="json",
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data["char_field"], "TESTChar2")
        self.assertEqual(response.data["email"], "changetest@iteo.com")

    def test_example_update_action_put(self):
        data = self.data
        data["char_field"] = "TESTChar2"
        response = self.client.put(
            reverse(
                "permissionless-model-examples-example-update-action", kwargs={"pk": 0}
            ),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data["char_field"], "TESTChar2")

    def test_example_non_detail_many_action(self):
        response = self.client.get(
            reverse("permissionless-model-examples-example-many-test-action"),
            format="json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 0)

    def test_car_detail_create_no_tag_names_failure(self):
        m = Manufacturer.objects.create(name="Volov", slug_name="volvo")
        response = self.client.post(
            reverse("car_edit_viewset-list"),
            data={
                "name": "SomeName",
                "body_type": 1,
                "engine_size": 1.9,
                "engine_type": 2,
                "manufacturer": m.id,
                "tags": [{"name": None}, {"name": None}],
            },
            format="json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data,
            {
                "errors": {
                    "tags": [
                        {
                            "name": [
                                ErrorDetail(
                                    string="This field may not be null.", code="null"
                                )
                            ]
                        },
                        {
                            "name": [
                                ErrorDetail(
                                    string="This field may not be null.", code="null"
                                )
                            ]
                        },
                    ]
                }
            },
        )

    def test_car_detail_create_success(self):
        m = Manufacturer.objects.create(name="Volov", slug_name="volvo")
        response = self.client.post(
            reverse("car_edit_viewset-list"),
            data={
                "name": "SomeName",
                "body_type": 1,
                "engine_size": 1.9,
                "engine_type": 2,
                "manufacturer": m.id,
                "tags": [{"name": "Tag"}],
            },
            format="json",
        )
        self.assertEqual(response.status_code, 201)
        self.assertDictEqual(
            response.data,
            {
                "id": 1,
                "name": "SomeName",
                "body_type": 1,
                "manufacturer": 1,
                "engine_size": 1.9,
                "engine_type": 2,
                "tags": [OrderedDict([("name", "Tag")])],
            },
        )

    def test_car_detail_update_no_tag_names_failure(self):
        m = Manufacturer.objects.create(name="Volov", slug_name="volvo")
        car = Car.objects.create(
            **{
                "name": "SomeName",
                "body_type": 1,
                "engine_size": 1.9,
                "engine_type": 2,
                "manufacturer": m,
            }
        )
        response = self.client.put(
            reverse("car_edit_viewset-detail", kwargs={"id": car.id}),
            data={
                "name": "SomeName",
                "body_type": 1,
                "engine_size": 1.9,
                "engine_type": 2,
                "manufacturer": m.id,
                "tags": [{"name": None}, {"name": None}],
            },
            format="json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data,
            {
                "errors": {
                    "tags": [
                        {
                            "name": [
                                ErrorDetail(
                                    string="This field may not be null.", code="null"
                                )
                            ]
                        },
                        {
                            "name": [
                                ErrorDetail(
                                    string="This field may not be null.", code="null"
                                )
                            ]
                        },
                    ]
                }
            },
        )

    def test_car_detail_update_success(self):
        m = Manufacturer.objects.create(name="Volov", slug_name="volvo")
        car = Car.objects.create(
            **{
                "name": "SomeName",
                "body_type": 1,
                "engine_size": 1.9,
                "engine_type": 2,
                "manufacturer": m,
            }
        )
        response = self.client.put(
            reverse("car_edit_viewset-detail", kwargs={"id": car.id}),
            data={
                "name": "SomeName",
                "body_type": 1,
                "engine_size": 1.9,
                "engine_type": 2,
                "manufacturer": m.id,
                "tags": [{"name": "Tag"}],
            },
            format="json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(
            response.data,
            {
                "id": 1,
                "name": "SomeName",
                "body_type": 1,
                "manufacturer": 1,
                "engine_size": 1.9,
                "engine_type": 2,
                "tags": [OrderedDict([("name", "Tag")])],
            },
        )
